/*****************************************************************************
 * XML lexer functions.
 * Copyright (C) 2004 Joe Wingbermuehle
 *****************************************************************************/

#include "jwm.h"

static const int BLOCK_SIZE = 16;

static TokenNode *head, *current;

static TokenNode *CreateNode(TokenNode *parent);
static AttributeNode *CreateAttribute(TokenNode *np); 

static int IsElementEnd(char ch);
static int IsValueEnd(char ch);
static int IsAttributeEnd(char ch);
static int IsSpace(char ch);
static char *ReadElementName(const char *line);
static char *ReadElementValue(const char *line);
static char *ReadAttributeValue(const char *line);
static int ParseEntity(const char *entity, char *ch);

static TokenType LookupType(const char *name);
static TokenType LookupTypeA(const char *name);
static TokenType LookupTypeB(const char *name);
static TokenType LookupTypeD(const char *name);
static TokenType LookupTypeE(const char *name);
static TokenType LookupTypeF(const char *name);
static TokenType LookupTypeH(const char *name);
static TokenType LookupTypeJ(const char *name);
static TokenType LookupTypeK(const char *name);
static TokenType LookupTypeL(const char *name);
static TokenType LookupTypeM(const char *name);
static TokenType LookupTypeO(const char *name);
static TokenType LookupTypeP(const char *name);
static TokenType LookupTypeR(const char *name);
static TokenType LookupTypeS(const char *name);
static TokenType LookupTypeT(const char *name);
static TokenType LookupTypeW(const char *name);

/*****************************************************************************
 *****************************************************************************/
TokenNode *Tokenize(const char *line) {
	TokenNode *np;
	AttributeNode *ap;
	char *temp;
	int inElement;
	int x;
	int found;

	head = NULL;
	current = NULL;
	inElement = 0;

	x = 0;
	/* Skip any initial white space */
	while(IsSpace(line[x])) ++x;

	/* Skip any XML stuff */
	if(!strncmp(line + x, "<?", 2)) {
		while(line[x]) {
			if(!strncmp(line + x, "?>", 2)) {
				x += 2;
				break;
			}
			++x;
		}
	}

	while(line[x]) {

		do {

			while(IsSpace(line[x])) ++x;

			/* Skip comments */
			found = 0;
			if(!strncmp(line + x, "<!--", 4)) {
				while(line[x]) {
					if(!strncmp(line + x, "-->", 3)) {
						x += 3;
						found = 1;
						break;
					}
					++x;
				}
			}
		} while(found);

		switch(line[x]) {
		case '<':
			++x;
			if(line[x] == '/') {
				++x;
				temp = ReadElementName(line + x);

				if(current) {

					if(temp) {

						if(current->type != LookupType(temp)) {
							Warning("configuration: close tag \"%s\" does not "
								"match open tag \"%s\"", temp,
								GetTokenName(current->type));
						}

					} else {
						Warning("configuration: unexpected and invalid close tag");
					}

					current = current->parent;
				} else {
					if(temp) {
						Warning("configuration: close tag \"%s\" without open "
							"tag", temp);
					} else {
						Warning("configuration: invalid close tag");
					}
				}

				if(temp) {
					x += strlen(temp);
					Release(temp);
				}

			} else {
				np = current;
				current = NULL;
				np = CreateNode(np);
				temp = ReadElementName(line + x);
				if(temp) {
					x += strlen(temp);
					np->type = LookupType(temp);
					Release(temp);
				} else {
					Warning("configuration: invalid open tag");
				}
			}
			inElement = 1;
			break;
		case '/':
			if(inElement) {
				++x;
				if(line[x] == '>' && current) {
					++x;
					current = current->parent;
					inElement = 0;
				} else {
					Warning("configuration: invalid tag");
				}
			} else {
				goto ReadDefault;
			}
			break;
		case '>':
			++x;
			inElement = 0;
			break;
		default:
ReadDefault:
			if(inElement) {
				ap = CreateAttribute(current);
				ap->name = ReadElementName(line + x);
				if(ap->name) {
					x += strlen(ap->name);
					if(line[x] == '=') {
						++x;
					}
					if(line[x] == '\"') {
						++x;
					}
					ap->value = ReadAttributeValue(line + x);
					if(ap->value) {
						x += strlen(ap->value);
					}
					if(line[x] == '\"') {
						++x;
					}
				}
			} else {
				temp = ReadElementValue(line + x);
				if(temp) {
					x += strlen(temp);
					if(current) {
						if(current->value) {
							current->value = Reallocate(current->value,
								strlen(current->value) + strlen(temp) + 1);
							strcat(current->value, temp);
							Release(temp);
						} else {
							current->value = temp;
						}
					} else {
						if(temp[0]) {
							Warning("configuration: unexpected text: \"%s\"",
								temp);
						}
						Release(temp);
					}
				}
			}
			break;
		}
	}

	return head;
}

/*****************************************************************************
 * Parse an entity reference.
 * The entity value is returned in ch and the length of the entity
 * is returned as the value of the function.
 *****************************************************************************/
int ParseEntity(const char *entity, char *ch) {
	char *temp;
	int x;

	if(!strncmp("&quot;", entity, 6)) {
		*ch = '\"';
		return 6;
	} else if(!strncmp("&lt;", entity, 4)) {
		*ch = '<';
		return 4;
	} else if(!strncmp("&gt;", entity, 4)) {
		*ch = '>';
		return 4;
	} else if(!strncmp("&amp;", entity, 5)) {
		*ch = '&';
		return 5;
	} else if(!strncmp("&apos;", entity, 6)) {
		*ch = '\'';
		return 6;
	} else {
		for(x = 0; entity[x]; x++) {
			if(entity[x] == ';') {
				break;
			}
		}
		temp = Allocate(x + 2);
		strncpy(temp, entity, x + 1);
		temp[x + 1] = 0;
		Warning("configuration: invalid entity: \"%s\"", temp);
		Release(temp);
		*ch = 0;
		return 1;
	}
}

/*****************************************************************************
 *****************************************************************************/
int IsElementEnd(char ch) {
	switch(ch) {
	case ' ':
	case '\t':
	case '\n':
	case '\r':
	case '\"':
	case '>':
	case '<':
	case '/':
	case '=':
	case 0:
		return 1;
	default:
		return 0;
	}
}

/*****************************************************************************
 *****************************************************************************/
int IsAttributeEnd(char ch) {
	switch(ch) {
	case 0:
	case '\"':
		return 1;
	default:
		return 0;
	}
}

/*****************************************************************************
 *****************************************************************************/
int IsValueEnd(char ch) {
	switch(ch) {
	case 0:
	case '<':
		return 1;
	default:
		return 0;
	}
}

/*****************************************************************************
 *****************************************************************************/
int IsSpace(char ch) {
	switch(ch) {
	case ' ':
	case '\t':
	case '\n':
	case '\r':
		return 1;
	default:
		return 0;
	}
}

/*****************************************************************************
 *****************************************************************************/
char *ReadElementName(const char *line) {
	char *buffer;
	int len, max;
	int x;

	len = 0;
	max = BLOCK_SIZE;
	buffer = Allocate(max + 1);

	for(x = 0; !IsElementEnd(line[x]); x++) {
		buffer[len++] = line[x];
		if(len >= max) {
			max += BLOCK_SIZE;
			buffer = Reallocate(buffer, max + 1);
		}
	}
	buffer[len] = 0;

	return buffer;
}

/*****************************************************************************
 *****************************************************************************/
char *ReadElementValue(const char *line) {
	char *buffer;
	char ch;
	int len, max;
	int x;

	len = 0;
	max = BLOCK_SIZE;
	buffer = Allocate(max + 1);

	for(x = 0; !IsValueEnd(line[x]); x++) {
		if(line[x] == '&') {
			x += ParseEntity(line + x, &ch) - 1;
			if(ch) {
				buffer[len] = ch;
			} else {
				buffer[len] = line[x];
			}
		} else {
			buffer[len] = line[x];
		}
		++len;
		if(len >= max) {
			max += BLOCK_SIZE;
			buffer = Reallocate(buffer, max + 1);
		}
	}
	buffer[len] = 0;

	return buffer;
}

/*****************************************************************************
 *****************************************************************************/
char *ReadAttributeValue(const char *line) {
	char *buffer;
	char ch;
	int len, max;
	int x;

	len = 0;
	max = BLOCK_SIZE;
	buffer = Allocate(max + 1);

	for(x = 0; !IsAttributeEnd(line[x]); x++) {
		if(line[x] == '&') {
			x += ParseEntity(line + x, &ch) - 1;
			if(ch) {
				buffer[len] = ch;
			} else {
				buffer[len] = line[x];
			}
		} else {
			buffer[len] = line[x];
		}
		++len;
		if(len >= max) {
			max += BLOCK_SIZE;
			buffer = Reallocate(buffer, max + 1);
		}
	}
	buffer[len] = 0;

	return buffer;
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupType(const char *name) {

	Assert(name);

	switch(name[0]) {
	case 'A':
		return LookupTypeA(name);
	case 'B':
		return LookupTypeB(name);
	case 'D':
		return LookupTypeD(name);
	case 'E':
		return LookupTypeE(name);
	case 'F':
		return LookupTypeF(name);
	case 'H':
		return LookupTypeH(name);
	case 'J':
		return LookupTypeJ(name);
	case 'K':
		return LookupTypeK(name);
	case 'L':
		return LookupTypeL(name);
	case 'M':
		return LookupTypeM(name);
	case 'O':
		return LookupTypeO(name);
	case 'P':
		return LookupTypeP(name);
	case 'R':
		return LookupTypeR(name);
	case 'S':
		return LookupTypeS(name);
	case 'T':
		return LookupTypeT(name);
	case 'W':
		return LookupTypeW(name);
	default:
		return TOK_INVALID;
	}

}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeA(const char *name) {
	if(!strcmp(name, "ActiveBackground")) {
		return TOK_ACTIVEBACKGROUND;
	} else if(!strcmp(name, "ActiveForeground")) {
		return TOK_ACTIVEFOREGROUND;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeB(const char *name) {
	if(!strcmp(name, "Background")) {
		return TOK_BACKGROUND;
	} else if(!strcmp(name, "Border")) {
		return TOK_BORDER;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeD(const char *name) {
	if(!strcmp(name, "DesktopCount")) {
		return TOK_DESKTOPCOUNT;
	} else if(!strcmp(name, "DoubleClickSpeed")) {
		return TOK_DOUBLECLICKSPEED;
	} else if(!strcmp(name, "DoubleClickDelta")) {
		return TOK_DOUBLECLICKDELTA;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeE(const char *name) {
	if(!strcmp(name, "Exit")) {
		return TOK_EXIT;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeF(const char *name) {
	if(!strcmp(name, "Font")) {
		return TOK_FONT;
	} else if(!strcmp(name, "Foreground")) {
		return TOK_FOREGROUND;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeH(const char *name) {
	if(!strcmp(name, "Height")) {
		return TOK_HEIGHT;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeJ(const char *name) {
	if(!strcmp(name, "JWM")) {
		return TOK_JWM;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeK(const char *name) {
	if(!strcmp(name, "Key")) {
		return TOK_KEY;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeL(const char *name) {
	if(!strcmp(name, "Load")) {
		return TOK_LOAD;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeM(const char *name) {
	if(!strcmp(name, "Menu")) {
		return TOK_MENU;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeO(const char *name) {
	if(!strcmp(name, "Outline")) {
		return TOK_OUTLINE;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeP(const char *name) {
	if(!strcmp(name, "Pager")) {
		return TOK_PAGER;
	} else if(!strcmp(name, "Program")) {
		return TOK_PROGRAM;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeR(const char *name) {
	if(!strcmp(name, "Restart")) {
		return TOK_RESTART;
	} else if(!strcmp(name, "RootMenu")) {
		return TOK_ROOTMENU;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeS(const char *name) {
	if(!strcmp(name, "Separator")) {
		return TOK_SEPARATOR;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeT(const char *name) {
	if(!strcmp(name, "Tray")) {
		return TOK_TRAY;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenType LookupTypeW(const char *name) {
	if(!strcmp(name, "Width")) {
		return TOK_WIDTH;
	} else {
		return TOK_INVALID;
	}
}

/*****************************************************************************
 *****************************************************************************/
const char *GetTokenName(TokenType type) {
	switch(type) {
	case TOK_ACTIVEBACKGROUND:
		return "ActiveBackground";
	case TOK_ACTIVEFOREGROUND:
		return "ActiveForeground";
	case TOK_BACKGROUND:
		return "Background";
	case TOK_BORDER:
		return "Border";
	case TOK_DESKTOPCOUNT:
		return "DesktopCount";
	case TOK_DOUBLECLICKSPEED:
		return "DoubleClickSpeed";
	case TOK_DOUBLECLICKDELTA:
		return "DoubleClickDelta";
	case TOK_EXIT:
		return "Exit";
	case TOK_FONT:
		return "Font";
	case TOK_FOREGROUND:
		return "Foreground";
	case TOK_HEIGHT:
		return "Height";
	case TOK_JWM:
		return "JWM";
	case TOK_KEY:
		return "Key";
	case TOK_LOAD:
		return "Load";
	case TOK_MENU:
		return "Menu";
	case TOK_OUTLINE:
		return "Outline";
	case TOK_PAGER:
		return "Pager";
	case TOK_PROGRAM:
		return "Program";
	case TOK_RESTART:
		return "Restart";
	case TOK_ROOTMENU:
		return "RootMenu";
	case TOK_SEPARATOR:
		return "Separator";
	case TOK_TRAY:
		return "Tray";
	case TOK_WIDTH:
		return "Width";
	default:
		return "[invalid]";
	}
}

/*****************************************************************************
 *****************************************************************************/
TokenNode *CreateNode(TokenNode *parent) {
	TokenNode *np;

	np = Allocate(sizeof(TokenNode));
	np->type = TOK_INVALID;
	np->value = NULL;
	np->attributes = NULL;
	np->subnodeHead = NULL;
	np->subnodeTail = NULL;
	np->parent = parent;
	np->next = NULL;

	if(!head) {
		head = np;
	}
	if(parent) {
		if(parent->subnodeHead) {
			parent->subnodeTail->next = np;
		} else {
			parent->subnodeHead = np;
		}
		parent->subnodeTail = np;
	} else if(current) {
		current->next = np;
	}
	current = np;

	return np;
}

/*****************************************************************************
 *****************************************************************************/
AttributeNode *CreateAttribute(TokenNode *np) {
	AttributeNode *ap;

	ap = Allocate(sizeof(AttributeNode));
	ap->name = NULL;
	ap->value = NULL;

	ap->next = np->attributes;
	np->attributes = ap;

	return ap;
}


