/****************************************************************************
 * Parser for the JWM XML configuration file.
 * Copyright (C) 2004 Joe Wingbermuehle
 ****************************************************************************/

#include "jwm.h"

static const char *restartCommand = "#restart";
static const char *restartName = "Restart";
static const char *exitCommand = "#exit";
static const char *exitName = "Exit";

static TokenNode *tokens;

static char *ReadFile(FILE *fd);

static void Parse(TokenNode *start);

static void ParseRootMenu(TokenNode *start);
static void ParseMenuItem(TokenNode *start, MenuType *menu);
static void ParseKey(TokenNode *tp);
static void ParseBorder(TokenNode *tp);
static void ParseTray(TokenNode *tp);
static void ParsePager(TokenNode *tp);
static void ParseLoad(TokenNode *tp);
static void ParseMenu(TokenNode *tp);

static char *FindAttribute(AttributeNode *ap, const char *name);
static void ReleaseTokens(TokenNode *np);
static void SetDesktopCount(const char *value);

/****************************************************************************
 ****************************************************************************/
void ParseConfig(const char *fileName) {
	FILE *fd;
	char *buffer;

	/* Try custom configuration first */
	fd = fopen(fileName, "r");
	if(!fd) {
		/* Try system configuration */
		fd = fopen(SYSTEM_CONFIG, "r");
		if(!fd) {
			Warning("could not open configuration file: %s or %s",
				fileName, SYSTEM_CONFIG);
			return;
		}
	}

	buffer = ReadFile(fd);
	fclose(fd);

	tokens = Tokenize(buffer);
	Release(buffer);
	Parse(tokens);
	ReleaseTokens(tokens);
}

/***************************************************************************
 ***************************************************************************/
void ReleaseTokens(TokenNode *np) {
	AttributeNode *ap;
	TokenNode *tp;

	while(np) {
		tp = np->next;

		while(np->attributes) {
			ap = np->attributes->next;
			if(np->attributes->name) {
				Release(np->attributes->name);
			}
			if(np->attributes->value) {
				Release(np->attributes->value);
			}
			Release(np->attributes);
			np->attributes = ap;
		}

		if(np->subnodeHead) {
			ReleaseTokens(np->subnodeHead);
		}

		if(np->value) {
			Release(np->value);
		}

		Release(np);
		np = tp;
	}

}

/***************************************************************************
 ***************************************************************************/
void Parse(TokenNode *start) {
	TokenNode *tp;
	int parsedRootMenu;

	if(!start) {
		return;
	}

	if(start->type == TOK_JWM) {
		parsedRootMenu = 0;
		for(tp = start->subnodeHead; tp; tp = tp->next) {
			switch(tp->type) {
			case TOK_BORDER:
				ParseBorder(tp);
				break;
			case TOK_TRAY:
				ParseTray(tp);
				break;
			case TOK_PAGER:
				ParsePager(tp);
				break;
			case TOK_LOAD:
				ParseLoad(tp);
				break;
			case TOK_MENU:
				ParseMenu(tp);
				break;
			case TOK_DESKTOPCOUNT:
				SetDesktopCount(tp->value);
				break;
			case TOK_DOUBLECLICKSPEED:
				SetDoubleClickSpeed(tp->value);
				break;
			case TOK_DOUBLECLICKDELTA:
				SetDoubleClickDelta(tp->value);
				break;
			case TOK_ROOTMENU:
				if(!parsedRootMenu) {
					ParseRootMenu(tp);
					parsedRootMenu = 1;
				} else {
					Warning("configuration error: "
						"root menu defined multiple times");
				}
				break;
			case TOK_KEY:
				ParseKey(tp);
				break;
			default:
				Warning("configuration error: invalid token: %d", tp->type);
				break;
			}
		}
	} else {
		Warning("configuration error: invalid start token: %d", start->type);
	}

}

/****************************************************************************
 ****************************************************************************/
void ParseRootMenu(TokenNode *start) {
	MenuType *menu;

	menu = Allocate(sizeof(MenuType));
	ParseMenuItem(start->subnodeHead, menu);
	SetRootMenu(menu);

}

/****************************************************************************
 ****************************************************************************/
void ParseMenuItem(TokenNode *start, MenuType *menu) {
	MenuItemType *current, *last;
	char *value;

	menu->offsets = NULL;
	menu->items = NULL;
	current = NULL;
	last = NULL;
	while(start) {
		switch(start->type) {
		case TOK_MENU:

			current = Allocate(sizeof(MenuItemType));
			value = FindAttribute(start->attributes, "label");
			if(value) {
				current->name = Allocate(strlen(value) + 1);
				strcpy(current->name, value);
			} else {
				current->name = NULL;
			}
			current->command = NULL;
			current->next = NULL;
			current->submenu = NULL;
			if(last) {
				last->next = current;
			} else {
				menu->items = current;
			}
			last = current;

			current->submenu = Allocate(sizeof(MenuType));
			ParseMenuItem(start->subnodeHead, current->submenu);

			break;
		case TOK_PROGRAM:

			current = Allocate(sizeof(MenuItemType));
			value = FindAttribute(start->attributes, "label");
			if(value) {
				current->name = Allocate(strlen(value) + 1);
				strcpy(current->name, value);
			} else if(start->value) {
				current->name = Allocate(strlen(start->value) + 1);
				strcpy(current->name, start->value);
			} else {
				current->name = NULL;
			}
			if(start->value) {
				current->command = Allocate(strlen(start->value) + 1);
				strcpy(current->command, start->value);
			} else {
				current->command = NULL;
			}
			current->submenu = NULL;
			current->next = NULL;
			if(last) {
				last->next = current;
			} else {
				menu->items = current;
			}
			last = current;

			break;
		case TOK_SEPARATOR:

			current = Allocate(sizeof(MenuItemType));
			current->name = NULL;
			current->command = NULL;
			current->next = NULL;
			current->submenu = NULL;
			if(last) {
				last->next = current;
			} else {
				menu->items = current;
			}
			last = current;

			break;
		case TOK_EXIT:

			current = Allocate(sizeof(MenuItemType));
			current->name = Allocate(strlen(exitName) + 1);
			strcpy(current->name, exitName);
			current->command = Allocate(strlen(exitCommand) + 1);
			strcpy(current->command, exitCommand);
			current->next = NULL;
			current->submenu = NULL;
			if(last) {
				last->next = current;
			} else {
				menu->items = current;
			}
			last = current;

			break;
		case TOK_RESTART:

			current = Allocate(sizeof(MenuItemType));
			current->name = Allocate(strlen(restartName) + 1);
			strcpy(current->name, restartName);
			current->command = Allocate(strlen(restartCommand) + 1);
			strcpy(current->command, restartCommand);
			current->next = NULL;
			current->submenu = NULL;
			if(last) {
				last->next = current;
			} else {
				menu->items = current;
			}
			last = current;

			break;
		default:
			Warning("configuration error: invalid node in menu: %d",
				start->type);
			break;
		}
		start = start->next;
	}
}

/****************************************************************************
 ****************************************************************************/
void ParseKey(TokenNode *tp) {
	char *key;
	char *mask;
	char *action;
	KeyType k;

	mask = FindAttribute(tp->attributes, "mask");
	key = FindAttribute(tp->attributes, "key");
	if(key == NULL) {
		Warning("configuration error: no key specified for Key");
		return;
	}

	action = tp->value;
	if(action == NULL) {
		Warning("configuration error: no action specified for Key");
		return;
	}

	if(!strcmp(action, "up")) {
		k = KEY_UP;
	} else if(!strcmp(action, "down")) {
		k = KEY_DOWN;
	} else if(!strcmp(action, "right")) {
		k = KEY_RIGHT;
	} else if(!strcmp(action, "left")) {
		k = KEY_LEFT;
	} else if(!strcmp(action, "escape")) {
		k = KEY_ESC;
	} else if(!strcmp(action, "select")) {
		k = KEY_ENTER;
	} else if(!strcmp(action, "next")) {
		k = KEY_NEXT;
	} else if(!strcmp(action, "close")) {
		k = KEY_CLOSE;
	} else if(!strcmp(action, "minimize")) {
		k = KEY_MIN;
	} else if(!strcmp(action, "maximize")) {
		k = KEY_MAX;
	} else if(!strcmp(action, "shade")) {
		k = KEY_SHADE;
	} else if(!strcmp(action, "move")) {
		k = KEY_MOVE;
	} else if(!strcmp(action, "resize")) {
		k = KEY_RESIZE;
	} else if(!strcmp(action, "root")) {
		k = KEY_ROOT;
	} else if(!strcmp(action, "window")) {
		k = KEY_WIN;
	} else if(!strcmp(action, "desktop")
		|| !strcmp(action, "desktop#")) {
		k = KEY_DESKTOP;
	} else {
		Warning("configuration error: invalid action: \"%s\"", action);
		return;
	}

	InsertBinding(k, mask, key);

}

/***************************************************************************
 ***************************************************************************/
void ParseBorder(TokenNode *tp) {
	TokenNode *np;
	char *aa;

	for(np = tp->subnodeHead; np; np = np->next) {
		switch(np->type) {
		case TOK_FONT:
			aa = FindAttribute(np->attributes, "antialias");
			if(aa && !strcmp(aa, "false")) {
				SetFont(FONT_BORDER, np->value, 0);
			} else {
				SetFont(FONT_BORDER, np->value, 1);
			}
			break;
		case TOK_WIDTH:
			SetBorderWidth(np->value);
			break;
		case TOK_HEIGHT:
			SetTitleHeight(np->value);
			break;
		case TOK_FOREGROUND:
			SetColor(COLOR_BORDER_FG, np->value);
			break;
		case TOK_BACKGROUND:
			SetColor(COLOR_BORDER_BG, np->value);
			break;
		case TOK_ACTIVEFOREGROUND:
			SetColor(COLOR_BORDER_ACTIVE_FG, np->value);
			break;
		case TOK_ACTIVEBACKGROUND:
			SetColor(COLOR_BORDER_ACTIVE_BG, np->value);
			break;
		default:
			Warning("configuration error: invalid Border option");
			break;
		}
	}

}

/***************************************************************************
 ***************************************************************************/
void ParseTray(TokenNode *tp) {
	TokenNode *np;
	char *aa;

	for(np = tp->subnodeHead; np; np = np->next) {
		switch(np->type) {
		case TOK_HEIGHT:
			SetTrayHeight(np->value);
			break;
		case TOK_FONT:
			aa = FindAttribute(np->attributes, "antialias");
			if(aa && !strcmp(aa, "false")) {
				SetFont(FONT_TRAY, np->value, 0);
			} else {
				SetFont(FONT_TRAY, np->value, 1);
			}
			break;
		case TOK_FOREGROUND:
			SetColor(COLOR_TRAY_FG, np->value);
			break;
		case TOK_BACKGROUND:
			SetColor(COLOR_TRAY_BG, np->value);
			break;
		case TOK_ACTIVEFOREGROUND:
			SetColor(COLOR_TRAY_ACTIVE_FG, np->value);
			break;
		case TOK_ACTIVEBACKGROUND:
			SetColor(COLOR_TRAY_ACTIVE_BG, np->value);
			break;
		default:
			Warning("configuration error: invalid Tray option");
			break;
		}
	}

}

/***************************************************************************
 ***************************************************************************/
void ParsePager(TokenNode *tp) {
	TokenNode *np;

	for(np = tp->subnodeHead; np; np = np->next) {
		switch(np->type) {
		case TOK_OUTLINE:
			SetColor(COLOR_PAGER_OUTLINE, np->value);
			break;
		case TOK_FOREGROUND:
			SetColor(COLOR_PAGER_FG, np->value);
			break;
		case TOK_BACKGROUND:
			SetColor(COLOR_PAGER_BG, np->value);
			break;
		case TOK_ACTIVEFOREGROUND:
			SetColor(COLOR_PAGER_ACTIVE_FG, np->value);
			break;
		case TOK_ACTIVEBACKGROUND:
			SetColor(COLOR_PAGER_ACTIVE_BG, np->value);
			break;
		default:
			Warning("configuration error: invalid Pager option");
			break;
		}
	}

}

/***************************************************************************
 ***************************************************************************/
void ParseLoad(TokenNode *tp) {
	TokenNode *np;

	for(np = tp->subnodeHead; np; np = np->next) {
		switch(np->type) {
		case TOK_OUTLINE:
			SetColor(COLOR_LOAD_OUTLINE, np->value);
			break;
		case TOK_FOREGROUND:
			SetColor(COLOR_LOAD_FG, np->value);
			break;
		case TOK_BACKGROUND:
			SetColor(COLOR_LOAD_BG, np->value);
			break;
		default:
			Warning("configuration error: invalid Load option");
			break;
		}
	}

}

/***************************************************************************
 ***************************************************************************/
void ParseMenu(TokenNode *tp) {
	TokenNode *np;
	char *aa;

	for(np = tp->subnodeHead; np; np = np->next) {
		switch(np->type) {
		case TOK_FONT:
			aa = FindAttribute(np->attributes, "antialias");
			if(aa && !strcmp(aa, "false")) {
				SetFont(FONT_MENU, np->value, 0);
			} else {
				SetFont(FONT_MENU, np->value, 1);
			}
			break;
		case TOK_FOREGROUND:
			SetColor(COLOR_MENU_FG, np->value);
			break;
		case TOK_BACKGROUND:
			SetColor(COLOR_MENU_BG, np->value);
			break;
		case TOK_ACTIVEFOREGROUND:
			SetColor(COLOR_MENU_ACTIVE_FG, np->value);
			break;
		case TOK_ACTIVEBACKGROUND:
			SetColor(COLOR_MENU_ACTIVE_BG, np->value);
			break;
		default:
			Warning("configuration error: invalid Menu option");
			break;
		}
	}

}

/***************************************************************************
 ***************************************************************************/
char *FindAttribute(AttributeNode *ap, const char *name) {

	while(ap) {
		if(!strcmp(name, ap->name)) {
			return ap->value;
		}
		ap = ap->next;
	}

	return NULL;
}

/***************************************************************************
 ***************************************************************************/
char *ReadFile(FILE *fd) {

	const int BLOCK_SIZE = 16;

	char *buffer;
	int len, max;
	int ch;

	len = 0;
	max = BLOCK_SIZE;
	buffer = Allocate(max + 1);

	for(;;) {
		ch = fgetc(fd);
		if(ch == EOF) {
			break;
		}
		buffer[len++] = ch;
		if(len >= max) {
			max += BLOCK_SIZE;
			buffer = Reallocate(buffer, max + 1);
		}
	}
	buffer[len] = 0;

	return buffer;
}

/****************************************************************************
 ****************************************************************************/
void SetDesktopCount(const char *value) {
	desktopCount = atoi(value);
	if(desktopCount <= 0) {
		desktopCount = 0;
	}
	if(desktopCount > MAX_DESKTOP_COUNT) {
		desktopCount = MAX_DESKTOP_COUNT;
	}
}


