/***************************************************************************
 * Functions to handle the tray.
 * Copyright (C) 2004 Joe Wingbermuehle
 ***************************************************************************/

#include "jwm.h"

typedef struct TrayNode {
	ClientNode *client;
	struct TrayNode *next;
} TrayNode;

int trayHeight = DEFAULT_TRAY_HEIGHT;
static const int TRAY_START = 64;

static char *menuTitle = "JWM";

static char minimized_bitmap[] = {
	0x06, 0x0F,
	0x0F, 0x06
};

Window trayWindow;
static GC trayGC;
static Pixmap buffer; 
static GC bufferGC;
static Pixmap minimizedPixmap;
static unsigned int trayStop;
static unsigned int trayTextOffset;

static TrayNode *trayNodes;

static unsigned int GetItemCount();
static unsigned int GetItemWidth();
static int ShouldShowItemInTray(ClientNode *np);

/***************************************************************************
 ***************************************************************************/
void CreateTray() {
	XSetWindowAttributes attr;

	attr.override_redirect = True;
	attr.event_mask = ButtonPressMask | ExposureMask | KeyPressMask;
	attr.background_pixel = colors[COLOR_TRAY_BG];

	trayWindow = JXCreateWindow(display, rootWindow, 0,
		rootHeight - trayHeight, rootWidth, trayHeight, 0, rootDepth,
		InputOutput, DefaultVisual(display, rootScreen),
		CWOverrideRedirect | CWBackPixel | CWEventMask, &attr);
	JXMapRaised(display, trayWindow);

	SetDefaultCursor(trayWindow);

	trayGC = JXCreateGC(display, trayWindow, 0, NULL);

	buffer = JXCreatePixmap(display, rootWindow,
		rootWidth - 1, trayHeight, rootDepth);
	bufferGC = JXCreateGC(display, buffer, 0, 0);
	JXSetForeground(display, bufferGC, colors[COLOR_TRAY_BG]);
	JXFillRectangle(display, buffer, bufferGC, 0, 0, rootWidth - 1,
		trayHeight);

	trayStop = JXTextWidth(fonts[FONT_TRAY], "10:00 MM", 8) + LOAD_WIDTH + 5;

	trayTextOffset = (trayHeight / 2)
		- (fonts[FONT_TRAY]->ascent + fonts[FONT_TRAY]->descent) / 2;

	trayNodes = NULL;

	minimizedPixmap = JXCreatePixmapFromBitmapData(display, rootWindow,
		minimized_bitmap, 4, 4, colors[COLOR_TRAY_FG],
		colors[COLOR_TRAY_BG], rootDepth);

}

/***************************************************************************
 ***************************************************************************/
void AddClientToTray(ClientNode *np) {
	TrayNode *tp;

	tp = Allocate(sizeof(TrayNode));
	tp->client = np;
	tp->next = trayNodes;
	trayNodes = tp;

	DrawTray();
}

/***************************************************************************
 ***************************************************************************/
void RemoveClientFromTray(ClientNode *np) {
	TrayNode *tp, *lp;

	lp = NULL;
	for(tp = trayNodes; tp; tp = tp->next) {
		if(tp->client == np) {
			if(lp) {
				lp->next = tp->next;
			} else {
				trayNodes = tp->next;
			}
			Release(tp);
			break;
		}
		lp = tp;
	}

	DrawTray();
}

/***************************************************************************
 ***************************************************************************/
void FocusNext() {
	TrayNode *tp;

	for(tp = trayNodes; tp; tp = tp->next) {
		if(!(tp->client->statusFlags & STAT_STICKY)
			&& currentDesktop != tp->client->desktop) {
			continue;
		}
		if(tp->client->statusFlags & STAT_ACTIVE) {
			if(tp->next) {
				tp = tp->next;
			} else {
				tp = trayNodes;
			}
			break;
		}
	}
	if(!tp) {
		tp = trayNodes;
	}

	while(tp && (tp->client->desktop != currentDesktop)
		&& !(tp->client->statusFlags & STAT_STICKY)) {
		tp = tp->next;
	}
	if(!tp) {
		tp = trayNodes;
		while(tp && (tp->client->desktop != currentDesktop)
			&& !(tp->client->statusFlags & STAT_STICKY)) {
			tp = tp->next;
		}
	}

	if(tp) {
		RestoreClient(tp->client);
		FocusClient(tp->client);
	}

}

/***************************************************************************
 ***************************************************************************/
void DestroyTray() {

	JXFreeGC(display, trayGC);
	JXFreeGC(display, bufferGC);
	JXFreePixmap(display, buffer);

	JXFreePixmap(display, minimizedPixmap);

	JXDestroyWindow(display, trayWindow);
}

/***************************************************************************
 ***************************************************************************/
void HandleTrayClick(XButtonEvent *event) {
	TrayNode *tp;
	int x, y;
	int realx, realy;
	int index, stop;
	unsigned int width, remainder;

	JXAllowEvents(display, ReplayPointer, CurrentTime);

	GetMousePosition(&realx, &realy);

	y = realy - rootHeight + trayHeight + 1;
	x = realx;

	width = GetItemWidth(GetItemCount());

	if(y > 1) {

		if(x < TRAY_START) {
			ShowRootMenu(0, rootHeight - trayHeight);
			return;
		} else {
			x -= TRAY_START;
		}

		if(x < pagerWidth) {
			switch(event->button) {
			case Button1:
			case Button3:
				ChangeDesktop(x / (pagerWidth / desktopCount));
				break;
			case Button4:
				NextDesktop();
				break;
			case Button5:
				PreviousDesktop();
				break;
			default:
				break;
			}
			return;
		} else {
			x -= pagerWidth + 1;
		}

		remainder = rootWidth - TRAY_START - pagerWidth - 2
			- trayStop - (width * GetItemCount());

		index = 0;
		for(tp = trayNodes; tp; tp = tp->next) {
			if(ShouldShowItemInTray(tp->client)) {

				if(remainder) {
					stop = index + width + 1;
					--remainder;
				} else {
					stop = index + width;
				}

				if(x >= index && x < stop) {
					switch(event->button) {
					case Button1:
						if(tp->client->statusFlags & STAT_ACTIVE) {
							MinimizeClient(tp->client);
						} else {
							RestoreClient(tp->client);
							FocusClient(tp->client);
						}
						break;
					case Button3:
						ShowWindowMenu(tp->client, realx, realy);
						break;
					default:
						break;
					}
					DrawTray();
					break;
				}
				index = stop;
			}
		}
	}
}

/***************************************************************************
 ***************************************************************************/
void NextDesktop() {
	ChangeDesktop((currentDesktop + 1) % desktopCount);
}

/***************************************************************************
 ***************************************************************************/
void PreviousDesktop() {
	if(currentDesktop > 0) {
		ChangeDesktop(currentDesktop - 1);
	} else {
		ChangeDesktop(desktopCount - 1);
	}
}

/***************************************************************************
 ***************************************************************************/
void ChangeDesktop(int desktop) {
	ClientNode *np;
	int x;

	if(desktop >= desktopCount || desktop < 0) {
		return;
	}

	if(currentDesktop == desktop && !initializing) {
		return;
	}

	for(x = 0; x < LAYER_COUNT; x++) {
		for(np = nodes[x]; np; np = np->next) {
			if(np->statusFlags & STAT_STICKY) {
				continue;
			}
			if(np->desktop == desktop) {
				ShowClient(np);
			} else if(np->desktop == currentDesktop) {
				HideClient(np);
			}
		}
	}

	currentDesktop = desktop;

	SetCardinalAtom(rootWindow, ATOM_NET_CURRENT_DESKTOP, currentDesktop);
	SetCardinalAtom(rootWindow, ATOM_WIN_WORKSPACE, currentDesktop);

	RestackClients();

	DrawTray();
}

/***************************************************************************
 ***************************************************************************/
void DrawTray() {
	TrayNode *tp;
	ButtonType buttonType;
	unsigned int width, x;
	unsigned int remainder;
	unsigned int itemCount;

	if(shouldExit) {
		return;
	}

	JXSetForeground(display, bufferGC, colors[COLOR_TRAY_BG]);
	JXFillRectangle(display, buffer, bufferGC, 0, 0,
		rootWidth - trayStop - 1, trayHeight);

	SetButtonDrawable(buffer, bufferGC);
	SetButtonSize(TRAY_START - 4, trayHeight - 3);
	SetButtonFont(FONT_TRAY);
	SetButtonAlignment(ALIGN_CENTER);
	DrawButton(1, 1, BUTTON_TRAY, menuTitle);

	itemCount = GetItemCount();

	if(!itemCount) {
		UpdatePager();
		UpdateTime();
		return;
	}

	width = GetItemWidth(itemCount);
	remainder = rootWidth - TRAY_START - pagerWidth - 2
		- trayStop - (width * itemCount);

	SetButtonAlignment(ALIGN_LEFT);

	x = TRAY_START + pagerWidth + 2;
	for(tp = trayNodes; tp; tp = tp->next) {

		if(!ShouldShowItemInTray(tp->client)) {
			continue;
		}

		if(tp->client->statusFlags & STAT_ACTIVE) {
			buttonType = BUTTON_TRAY_ACTIVE;
		} else {
			buttonType = BUTTON_TRAY;
		}

		if(remainder) {
			SetButtonSize(width - 2, trayHeight - 3);
		} else {
			SetButtonSize(width - 3, trayHeight - 3);
		}

		DrawButton(x, 1, buttonType, tp->client->name);

		if(tp->client->statusFlags & STAT_MINIMIZED) {
			JXCopyArea(display, minimizedPixmap, buffer, bufferGC, 0, 0,
				4, 4, x + 3, trayHeight - 8);
		}

		if(remainder) {
			x += width + 1;
			--remainder;
		} else {
			x += width;
		}

	}

	UpdatePager();
	UpdateTime();

}

/***************************************************************************
 ***************************************************************************/
void UpdatePager() {
	JXCopyArea(display, GetPager(), buffer, bufferGC,
		0, 0, pagerWidth, trayHeight, TRAY_START, 0);
	JXCopyArea(display, buffer, trayWindow, trayGC,
		TRAY_START, 0, pagerWidth, trayHeight, TRAY_START, 0);
}

/***************************************************************************
 ***************************************************************************/
void UpdateTime() {
	static char tstring[80];
	unsigned int len;
	time_t clock;
	struct tm *timeInfo;
	unsigned int x;

	clock = time(NULL);
	timeInfo = localtime(&clock);
	len = strftime(tstring, sizeof(tstring), "%I:%M %p", timeInfo);
	if(tstring[0] == '0') {
		for(x = 0; x < len; x++) {
			tstring[x] = tstring[x + 1];
		}
		--len;
	}
	x = rootWidth - JXTextWidth(fonts[FONT_TRAY], tstring, len) - 4;

	JXSetForeground(display, bufferGC, colors[COLOR_TRAY_BG]);
	JXFillRectangle(display, buffer, bufferGC,
		rootWidth - trayStop, 0, trayStop, trayHeight);

	RenderString(buffer, bufferGC, FONT_TRAY, RAMP_TRAY,
		x, trayTextOffset, rootWidth, tstring);

#ifdef SHOW_LOAD
	JXSetForeground(display, bufferGC, colors[COLOR_TRAY_DOWN]);
	JXDrawRectangle(display, buffer, bufferGC, rootWidth - trayStop,
		1, LOAD_WIDTH + 1, trayHeight - 3);
	JXCopyArea(display, GetLoadDisplay(), buffer, bufferGC, 0, 0,
		LOAD_WIDTH, trayHeight - 4, rootWidth - trayStop + 1, 2);
#endif

	JXCopyArea(display, buffer, trayWindow, trayGC,
		0, 0, rootWidth - 1, trayHeight, 0, 0);

}

/***************************************************************************
 ***************************************************************************/
unsigned int GetItemCount() {
	TrayNode *tp;
	unsigned int count;

	count = 0;
	for(tp = trayNodes; tp; tp = tp->next) {
		if(ShouldShowItemInTray(tp->client)) {
			++count;
		}
	}

	return count;
}

/***************************************************************************
 ***************************************************************************/
int ShouldShowItemInTray(ClientNode *np) {

	if(np->desktop != currentDesktop
		&& !(np->statusFlags & STAT_STICKY)) {
		return 0;
	}

	if(np->owner != None || (np->statusFlags & STAT_NOLIST)
		|| (np->statusFlags & STAT_WITHDRAWN)
		|| (!(np->statusFlags & STAT_MAPPED)
		&& !(np->statusFlags & (STAT_MINIMIZED | STAT_SHADED)))) {
		return 0;
	}

	return 1;
}

/***************************************************************************
 ***************************************************************************/
unsigned int GetItemWidth(unsigned int itemCount) {
	unsigned int width;

	width = rootWidth - TRAY_START - pagerWidth - 2 - trayStop;

	if(!itemCount) {
		return width;
	}

	width /= itemCount;
	if(!width) {
		width = 1;
	}

	return width;
}

/***************************************************************************
 ***************************************************************************/
void SetTrayHeight(const char *str) {
	int height;

	Assert(str);

	height = atoi(str);
	if(height < MIN_TRAY_HEIGHT || height > MAX_TRAY_HEIGHT) {
		trayHeight = DEFAULT_TRAY_HEIGHT;
		Warning("invalid tray height specified: %d", height);
	} else {
		trayHeight = height;
	}

}


